/*
 * Copyright (c) 2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package org.openjdk.jigsaw;

/**
 * <p> Architecture of a module-file, or module library </p>
 */

public class ModuleArchitecture {  // ## find a better name

    private static final String WILDCARD = "";

    public static final ModuleArchitecture ANY =
            new ModuleArchitecture(WILDCARD, WILDCARD);

    private final String os;
    private final String arch;

    private ModuleArchitecture(String os, String arch) {
        this.os = os == null ? WILDCARD : os;
        this.arch = arch == null ? WILDCARD : arch;
    }

    public static ModuleArchitecture create(String os,  String arch) {
        return new ModuleArchitecture(os, arch);
    }

    public static ModuleArchitecture createFromOS(String os) {
        return new ModuleArchitecture(os, WILDCARD);
    }

    public static ModuleArchitecture createFromArch(String arch) {
        return new ModuleArchitecture(WILDCARD, arch);
    }

    /**
     * Determines if the given ModuleArchitecture matches this.
     */
    public boolean matches(ModuleArchitecture modArch) {
        if (!equalOrAny(os, modArch.os))
            return false;
        if (!equalOrAny(arch, modArch.arch))
            return false;

        return true;
    }

    // case-sensitive comparison
    private boolean equalOrAny(String s1, String s2) {
        if (s1.equals(WILDCARD))
            return true;
        if (s2.equals(WILDCARD))
            return true;
        if (s1.equals(s2))
            return true;

        return false;
    }

    @Override
    public boolean equals(Object ob) {
        if (!(ob instanceof ModuleArchitecture))
            return false;
        ModuleArchitecture that = (ModuleArchitecture)ob;
        if (!(os.equals(that.os)))
            return false;
        if (!(arch.equals(that.arch)))
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        return (os.hashCode() * 41 + arch.hashCode());
    }

    public String os() {
        return os;
    }

    public String arch() {
        return arch;
    }

    @Override
    public String toString() {
        if (this.equals(ANY))
            return "ANY";
        if (arch.equals(WILDCARD))
            return os;
        if (os.equals(WILDCARD))
            return arch;
        return os + '-' + arch;
    }
}
